%%%-------------------------------------------------------------------
%%% @author Konstantin Kalin <kkalin@kkalin-macbook.local>
%%% @copyright (C) 2012, Konstantin Kalin
%%% @doc
%%%
%%% @end
%%% Created : 24 Apr 2012 by Konstantin Kalin <kkalin@kkalin-macbook.local>
%%%-------------------------------------------------------------------
-module(connections).

-behaviour(gen_server).

%% API
-export([start_link/1, start_link/2, get_connection/0, 
         show/0, stop/0]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

-define(SERVER, ?MODULE). 

-record(state, {
          connections,
          position = 0,
          urls,
          amount
         }).

%%%===================================================================
%%% API
%%%===================================================================

start_link(Urls) ->
    gen_server:start_link({local, ?SERVER}, ?MODULE, [Urls, 1], []).

start_link(Urls, Amount) ->
    gen_server:start_link({local, ?SERVER}, ?MODULE, [Urls, Amount], []).

get_connection() ->
    gen_server:call(?SERVER, get).

show() ->
    gen_server:cast(?SERVER, show).

stop() ->
    gen_server:cast(?SERVER, stop).

%%%===================================================================
%%% gen_server callbacks
%%%===================================================================

init([Urls, Amount]) ->
    %%process_flag(trap_exit, true),
    io:format("INIT --- Urls = ~p~n~n", [Urls]),
    {ok, #state{urls=Urls, amount=Amount}, 0}.

handle_call(get, _From, #state{connections=Connections, position=Pos} = State) ->
    Connection = array:get(Pos, Connections),
    NewPos = (Pos + 1) rem array:size(Connections),
    {reply, Connection, State#state{position=NewPos}}.

handle_cast(show, #state{connections=Connections}=State) ->
    io:format("Connections=[~p]~n", [array:to_list(Connections)]),
    {noreply, State};
handle_cast(stop, State) ->
    {stop, normal, State}.

handle_info(timeout, #state{urls=Urls, amount=1}=State) ->
    {ok, Connections} = init_connections(Urls),
    {noreply, State#state{connections=Connections}};
handle_info(timeout, #state{urls=Urls, amount=Amount}=State) ->
    {ok, Connections} = init_connections(Urls, Amount),
    {noreply, State#state{connections=Connections}}.

terminate(_Reason, #state{connections=Connections}) ->
    L = array:to_list(Connections),
    [close_connection(X) || X <- L],
    ok.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

init_connections(Urls, Amount) ->
    Conn = lists:flatten([init_connections(X, Amount, []) || X <- Urls]),
    {ok, array:from_list(Conn)}.

init_connections(Urls) ->
    Connections = array:from_list([init_connection(Url) || Url <- Urls]),
    {ok, Connections}.

init_connections(_Url, 0, Acc) ->
    Acc;
init_connections(Url, Amount, Acc) ->
    NewAcc = [init_connection(Url) | Acc],
    init_connections(Url, Amount - 1, NewAcc).

init_connection(Url) when is_atom(Url) ->
    init_connection(atom_to_list(Url));
init_connection(Url) ->
    io:format("Url = ~p~n", [Url]),
    {ok, Params} = amqp_uri:parse(Url),
    {ok, Connection} = amqp_connection:start(Params),
    Connection.

close_connection(Connection) ->
    io:format("Stopping connection: ~p~n", [Connection]),
    amqp_connection:close(Connection).
