%%%-------------------------------------------------------------------
%%% @author Konstantin Kalin <kkalin@kkalin-macbook>
%%% @copyright (C) 2012, Konstantin Kalin
%%% @doc
%%%
%%% @end
%%% Created : 20 Apr 2012 by Konstantin Kalin <kkalin@kkalin-macbook>
%%%-------------------------------------------------------------------
-module(consumer_wrk).

-behaviour(gen_server).

-include_lib("amqp_client/include/amqp_client.hrl").

%% API
-export([start_link/4, stop/1]).

%% gen_server callbacks
-export([init/1, handle_cast/2, handle_info/2,
         terminate/2]).

-define(SERVER, ?MODULE). 

-record(state, {
          queue :: binary(),
          channel,
          routing_key,
          exchange, 
          type,
          tag,
          connection,
          delivery_time = 0
         }).

%%%===================================================================
%%% API
%%%===================================================================

start_link(Connection, Exchange, Type, RoutingKey) ->
    gen_server:start_link(?MODULE, [Connection, Exchange, Type, RoutingKey], []).

stop(Pid) ->
    gen_server:cast(Pid, stop).

init([Connection, Exchange, Type, RoutingKey]) ->
    {ok, #state{connection=Connection, exchange=Exchange, 
                routing_key=RoutingKey, type=Type}, 0}.

handle_cast(stop, State) ->
    {stop, normal, State}.

init_channel_exchange(Connection, Exchange, Type) ->
    {ok, Channel} = amqp_connection:open_channel(Connection),
    Declare = #'exchange.declare'{exchange = Exchange, 
                                  type = Type, 
                                  nowait = true},
    Msg = amqp_channel:call(Channel, Declare),
    io:format("Msg=~p~n", [Msg]),
    {channel, Channel}.

init_queue_and_bind(Channel, Exchange, Key) ->
    QDeclare = #'queue.declare'{ 
      arguments = [{<<"x-expires">>, signedint, 1000}]
     },
    #'queue.declare_ok'{queue=Queue} = amqp_channel:call(Channel, QDeclare),

    Binding = #'queue.bind'{ 
      exchange = Exchange,
      queue = Queue,
      routing_key = Key
     },
    #'queue.bind_ok'{} = amqp_channel:call(Channel, Binding),

    Consume = #'basic.consume'{
      queue = Queue,
      no_ack = true
     },
    #'basic.consume_ok'{consumer_tag = Tag} = amqp_channel:call(Channel, Consume),
    {ok, Queue, Tag}.

handle_info(timeout, #state{queue=undefined, exchange=Exchange,
                            routing_key = Key, type=Type, 
                            connection=Connection} = State) ->
    {channel, Channel} = init_channel_exchange(Connection, Exchange, Type),
    {ok, Queue, Tag} = init_queue_and_bind(Channel, Exchange, Key),
    {noreply, State#state{channel=Channel, tag = Tag, queue = Queue}};

handle_info({#'basic.deliver'{delivery_tag=Tag}, 
             #'amqp_msg'{payload=Payload} = _Content},
            #state{channel=_Channel, delivery_time=Time} = State) ->
    %%amqp_channel:cast(Channel, #'basic.ack'{delivery_tag = Tag}),
    %%{_, _, Micro} = erlang:now(),
    %%io:format("Got message = ~p; Tag=~p; Now=~p~n", [Payload, Tag, Micro]),
    case Payload of
        <<"Start">> ->
            io:format("Start calculations~n"),
            {noreply, State#state{delivery_time = 0}};
        <<"Stop">> ->
            io:format("~p: avg delivery time=~p~n", [self(), Time]),
            {noreply, State};
        Value ->
            SendMicro = list_to_integer(binary_to_list(Value)),
            {value, NewTime} = calc_time(Time, SendMicro),
            {noreply, State#state{delivery_time = NewTime}}
        end;

handle_info({'basic.consume_ok', _Tag}, State) ->
    {noreply, State}.

terminate(_Reason, #state{channel = Channel}) ->
    amqp_channel:close(Channel),
    ok.

calc_time(Acc, SendTime) ->
    {_, _, RecTime} = erlang:now(),
    calc_time(Acc, SendTime, RecTime).

calc_time(Acc, SendTime, RecTime) when RecTime > SendTime ->
    NewAcc = Acc * 0.99 + ((RecTime - SendTime) * 0.01),
    {value, NewAcc};
calc_time(Acc, SendTime, RecTime) ->
    NewAcc = Acc * 0.99 + (((RecTime + 1000000) - SendTime) * 0.01),
    {value, NewAcc}.
