%%%-------------------------------------------------------------------
%%% @author Konstantin Kalin <kkalin@kkalin-macbook.local>
%%% @copyright (C) 2012, Konstantin Kalin
%%% @doc
%%%
%%% @end
%%% Created : 23 Apr 2012 by Konstantin Kalin <kkalin@kkalin-macbook.local>
%%%-------------------------------------------------------------------
-module(publisher_wrk).

-behaviour(gen_server).

-include_lib("amqp_client/include/amqp_client.hrl").

%% API
-export([start_link/6]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

-define(SERVER, ?MODULE). 

-record(state, {
          exchange :: binary(),
          type,
          key :: binary(),
          count :: integer(),
          rate :: integer(),
          channel,
          connection
         }).

%%%===================================================================
%%% API
%%%===================================================================
start_link(Connection, Exchange, Type, Key, Count, Rate) ->
    gen_server:start_link(?MODULE, {Connection, Exchange, Type, Key, Count, Rate}, []).

%%%===================================================================
%%% gen_server callbacks
%%%===================================================================

init({Connection, Exchange, Type, Key, Count, Rate}) ->
    {ok, #state{connection = Connection, exchange = Exchange, 
                key = Key, count = Count, type = Type, rate = Rate}, 0}.

handle_call(_Request, _From, State) ->
    Reply = ok,
    {reply, Reply, State}.

handle_cast(_Msg, State) ->
    {noreply, State}.

handle_info(timeout, #state{exchange = Exchange, connection = Connection,
                            rate = Rate, type = Type, key=Key} = State) ->
    io:format("Initializing a connection and channel to RabbitMQ~n"),
    {ok, Channel} =  amqp_connection:open_channel(Connection),
    Declare = #'exchange.declare'{exchange = Exchange, 
                                  type = Type, 
                                  nowait = true},
    ok = amqp_channel:call(Channel, Declare),
    io:format("Start sending messages with rate=[~p] per 1 sec~n", 
              [Rate]),
    {ok, Delay, _} = get_delay_count(Rate),
    erlang:send_after(Delay, self(), publish),
    publish_message(Channel, Exchange, Key, <<"Start">>),
    {noreply, State#state{channel = Channel, connection = Connection}};

handle_info(publish, #state{count = 0, channel=Channel, exchange=Exchange, 
                           key=Key} = State) ->
    publish_message(Channel, Exchange, Key, <<"Stop">>),
    io:format("Publishing has been finished~n"),
    {stop, normal, State};
handle_info(publish, #state{exchange = Exchange, key = Key, count = Count, 
                            rate = Rate, channel = Channel} = State) ->
    case get_delay_count(Rate) of 
        {ok, Delay, 1} ->
            publish_message(Channel, Exchange, Key, nowmicro_as_binary()),
            erlang:send_after(Delay, self(), publish);
        {ok, 0,  Amount} ->
            publish_messages(Channel, Exchange, Key, Amount),
            erlang:send_after(0, self(), publish)
    end,
    {noreply, State#state{count = (Count - 1)}}.

terminate(_Reason, #state{channel = Channel}) when Channel =/= undefined->
    io:format("Closing channel [~p]~n", [Channel]),
    amqp_channel:close(Channel),
    ok;
terminate(_Reason, _State) ->
    ok.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

nowmicro_as_binary() ->
    {_, _, Micro} = erlang:now(),
    list_to_binary(integer_to_list(Micro)).

publish_messages(_Channel, _Exchange, _Key, 0) ->
    ok;
publish_messages(Channel, Exchange, Key, Count) ->
    publish_message(Channel, Exchange, Key, nowmicro_as_binary()),
    publish_messages(Channel, Exchange, Key, (Count - 1)).

publish_message(Channel, Exchange, Key, Payload) when is_binary(Payload) -> 
    Publish = #'basic.publish' {
      exchange = Exchange,
      routing_key = Key
     },
    Msg = #'amqp_msg'{
      payload = Payload
     },
    amqp_channel:cast(Channel, Publish, Msg).    

get_delay_count(Rate) when Rate > 1000 ->
    {ok, 0, round(Rate/1000)};
get_delay_count(Rate) ->
    {ok, round(1000/Rate), 1}.
